/*************************************************************************************
 *
 *  vector.h:  class for vector
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#ifndef _JSMATH_VECTOR_H
#define _JSMATH_VECTOR_H 

#include "matrix.h"

#include <float.h>

namespace Math
{
	template<typename _T>class Quaternion;

	template<typename _T = double, const int dimensions = 3>
	class Vector: public Matrix<_T>
	{
	public:	

		Vector(): Matrix<_T>(dimensions, 1) {}
		
		Vector(_T comp1, _T comp2, _T comp3): Matrix<_T>(3, 1)
		{
			Create( comp1, comp2, comp3 );
		}

		void Create(_T comp1, _T comp2, _T comp3)
		{
			Matrix<_T>::m_values[0] = comp1;
			Matrix<_T>::m_values[1] = comp2;
			Matrix<_T>::m_values[2] = comp3;
		}

		//  Math::Vector<> position = 0;
		//  eg. in object declaration where i want to assign default value
		Vector( _T num ): Matrix<_T>(dimensions, 1)
		{
			operator=(num);
		}

		//  assignment operators are not inherited by derived classes
		Vector<_T> & operator=( const Vector<_T>& right_side )
		{	// deep-copy
			CHECK_SIZES(*this, right_side)

			if ( this == &right_side ) return *this;
			Assign(*this, right_side);

			return *this;
		}

		//  assignment operators are not inherited by derived classes
		Vector<_T> & operator=( const Matrix<_T> & right_side )
		{	// deep-copy
			if ( this == &right_side ) return *this;
			Assign(*this, right_side);

			return *this;
		}

		//  Math::Vector<><double> vec;
		//  vec = 1;		//  all components of vector set to 1
		//
		void operator=( _T num )
		{
			for (int i=0; i < Matrix<_T>::m_size; i++) 
				Matrix<_T>::m_values[i] = num;
		}

		Vector<_T> & operator=( const Quaternion<_T> & rightQuat )
		{
			Assign( *this, rightQuat.GetVector() );
			return *this;
		}


		//  Get magnitude (or length, norma, modul, absolut value)
		_T GetLength() const
		{
			_T result = 0;

			for(int i = 0; i < Matrix<_T>::m_size; i++)
				result += pow( Matrix<_T>::m_values[i], 2 );

			return sqrt(result);
		}

		// vector = scalar * vector;
		friend Vector<_T> operator*( const double & scalar, const Vector<_T> & vec )
		{
			Vector<_T> result;
			for(int i = 0; i < vec.m_size; i++)
				result.m_values[i] = scalar * vec.m_values[i];
			return result;
		}

		// vector = vector * scalar;
		friend Vector<_T> operator*( const Vector<_T> & vec, const _T & scalar )
		{
			Vector result;
			for(int i = 0; i < vec.m_size; i++)
				result.m_values[i] = vec.m_values[i] * scalar;
			return result;
		}

		//  compute scalar (or dot) product (= left * right)
		friend _T scalar_product( const Vector<_T> & left, const Vector<_T> & right )
		{
			CHECK_SIZES(left, right)

			_T result = 0;
			for(int i = 0; i < left.m_size; i++)
				result += left.m_values[i] * right.m_values[i];
			
			return result;
		}

		//  compute vector (or cross) product (= left x right)
		friend Vector<_T> vector_product( const Vector<_T> & left, const Vector<_T> & right )
		{
			Vector<_T> result;
			
			result.m_values[0] = left.m_values[1]*right.m_values[2] - left.m_values[2]*right.m_values[1];
			result.m_values[1] = left.m_values[2]*right.m_values[0] - left.m_values[0]*right.m_values[2];
			result.m_values[2] = left.m_values[0]*right.m_values[1] - left.m_values[1]*right.m_values[0];
			
			return result;
		}

		friend Vector<_T> operator+( const Vector<_T> & left, const Vector<_T> & right )
		{
			CHECK_SIZES(left, right)
			
			Vector<_T> result;
			for (int i=0; i < right.m_size; i++) result[i] = left.m_values[i] + right.m_values[i];
			
			return result;
		}

		friend Vector<_T> operator-( const Vector<_T> & left, const Vector<_T> & right )
		{
			CHECK_SIZES(left, right)

			Vector<_T> result;
			for (int i=0; i < right.m_size; i++) result[i] = left.m_values[i] - right.m_values[i];
			
			return result;
		}

		// vector = vector / scalar;
		friend Vector<_T> operator/( const Vector<_T> & vec, const _T & scalar )
		{
			Vector<_T> result;
			for(int i = 0; i < vec.m_size; i++)
				result.m_values[i] = vec.m_values[i] / scalar;

			return result;
		}


		//  reduces any ordinary vector to a unit normalized vector, length = 1
		void Normalize()
		{
			// Calculate the length of the vector
			_T length = GetLength();

			//  Keep the program from blowing up by vectors whose length 
			//  may be calculated too close to zero.
			if (length <= FLT_EPSILON)		//  == 0
				return;

			//  Dividing each element by the length will result in a unit
			//  normalized vector.
			Matrix<_T>::m_values[0] /= length;
			Matrix<_T>::m_values[1] /= length;
			Matrix<_T>::m_values[2] /= length;
		}

		virtual ~Vector() {}

	};  //  class Vector

}  //  end namespace Math


#endif //  _JSMATH_VECTOR_H
