/*************************************************************************************
 *
 *  bh.h:  implementation of Barnes-Hut algorithm
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#ifndef _JSSIMULATION_BH_H
#define _JSSIMULATION_BH_H

#include "../data/octtree.h"

#include <fstream>
#include <stdio.h>

namespace JakubSchwarzmeier
{
	namespace Simulations
	{
		namespace BarnesHut
		{
			class fstream
			{
			public:

				fstream()
				{
					m_file = NULL;
				}

				~fstream()
				{
					if (m_file != NULL) fclose(m_file);
				}

				void open(const char * szFileName, std::ios_base::openmode mode )
				{
					m_file = ::fopen( szFileName, "r+b" );
				}

				void seekg( long long pos, std::ios_base::seekdir way = std::ios_base::beg )
				{
					int origin;
					if (way == std::ios_base::beg)
						origin = SEEK_SET;
					else
					if (way == std::ios_base::end)
						origin = SEEK_END;
					else
						origin = SEEK_CUR;
					
					long long ll = _fseeki64( m_file, pos, origin);
					#ifdef _DEBUG
						if (ll != 0) throw;
					#endif
				}

				long long tellg()
				{
					return _ftelli64( m_file );
				}

				void close()
				{
					fclose(m_file);
					m_file = NULL;
				}

				FILE * m_file;

			};

			class ICommonNode: public Data::ICommonNode
			{
			public:
				
				ICommonNode( Math::Vector<> n_position = 0, double n_mass = 0)
				{
					m_position = n_position;
					m_mass = n_mass;
					m_load = 0;
				}

				virtual ~ICommonNode() {}

				Math::Vector<> m_position;
				double m_mass;
				double m_load;	//  costzones
			};
			
			class DataReader
			{
			public:

				DataReader()
				{
					m_stream = NULL;
					m_file = NULL;
				}

				virtual ~DataReader()
				{
					if (m_stream != NULL)  delete m_stream;
				}

				void OpenStream( std::fstream * stream )
				{
					m_stream = new std::istream( stream->rdbuf() );
				}

				void OpenStream( BarnesHut::fstream * stream )
				{
					m_file = stream->m_file;
				}

				void Read(double * d)
				{
					if (m_stream != NULL)
					{
						m_stream->read( reinterpret_cast<char*>(d), sizeof(*d) );
						#ifdef _DEBUG
						if (m_stream->fail())  throw;
						#endif
					}
					else
					{
						size_t x = fread( d, sizeof(*d), 1, m_file );
						#ifdef _DEBUG
						if (x == 0)  throw;
						#endif
					}
				}

				void close()
				{
					if (m_stream != NULL)
					{
						delete m_stream;
						m_stream = NULL;
					}
				}

			protected:

				std::istream * m_stream;
				FILE * m_file;
			};

			class DataWriter
			{
			public:

				DataWriter()
				{
					m_stream = NULL;
					m_file = NULL;
				}

				virtual ~DataWriter()
				{
					if (m_stream != NULL)  delete m_stream;
				}

				void OpenStream( std::ostream * stream )
				{
					m_stream = new std::ostream( stream->rdbuf() );
					m_stream->seekp( m_stream->tellp(), std::ios_base::beg );
				}

				void OpenStream( BarnesHut::fstream * stream )
				{
					m_file = stream->m_file;
				}

				void Write(double * d)
				{
					if (m_stream != NULL)
					{
						m_stream->write( reinterpret_cast<char*>(d), sizeof(*d) );
						#ifdef _DEBUG
						if ( m_stream->fail() )	throw;
						#endif
					}
					else
					{
						size_t ll = _fseeki64( m_file, _ftelli64( m_file ), SEEK_SET);

						size_t x = fwrite(d, sizeof(*d), 1, m_file);

						#ifdef _DEBUG
						if (x == 0)  throw;
						#endif
					}
				}

			protected:

				std::ostream * m_stream;
				FILE * m_file;
			};

			class Leaf: public ICommonNode
			{
			public:
				
				Leaf(): ICommonNode() 
				{						
					m_acceleration = 0;
				}

				Leaf(Math::Vector<> n_position, Math::Vector<> n_velocity, double n_mass): ICommonNode(n_position, n_mass)
				{
					m_velocity = n_velocity;
					m_acceleration = 0;
				}

				void Read(DataReader * dr)
				{
					dr->Read( &m_mass );

					dr->Read( &m_position[0] );
					dr->Read( &m_position[1] );
					dr->Read( &m_position[2] );

					dr->Read( &m_velocity[0] );
					dr->Read( &m_velocity[1] );
					dr->Read( &m_velocity[2] );
				}
				
				void Write(DataWriter * dw)
				{
					dw->Write( &m_mass );

					dw->Write( &m_position[0] );
					dw->Write( &m_position[1] );
					dw->Write( &m_position[2] );

					dw->Write( &m_velocity[0] );
					dw->Write( &m_velocity[1] );
					dw->Write( &m_velocity[2] );
				}

				virtual ~Leaf()
				{
				}

				Math::Vector<> m_velocity;
				Math::Vector<> m_acceleration;		//  neni treba, pocita se vzdy od znovu; u distributed obsahuje zrychleni udelene pouze vzdalenymi nody
			};

			class Node: public Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf >
			{
			public:

				Node(Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf > * n_ptrParent, Math::Vector<> * n_ptrCorner, double n_bisect, Math::Quaternion<> * n_ptrHilbertOrientation):
					Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf >(n_ptrParent, n_ptrCorner, n_bisect, n_ptrHilbertOrientation)
				{}

				virtual ~Node() {}

			protected:

			};

			class Group
			{
			public:

				ICommonNode * m_element;	//  particle or node
				Math::Vector<> m_acceleration, m_velocity, m_position;	//  far forces
				std::vector<ICommonNode*> m_nearList;	//  near forces

				Group( ICommonNode * n_element = NULL)
				{
					m_element = n_element;
					//  must explicitly set up all the matrices to zeros
					m_acceleration = 0;
					m_velocity = 0;
					m_position = 0;
				}
			};


			class Tree: public Data::OctTree<ICommonNode, Leaf, Node>
			{
			public:

				Tree( std::vector<Leaf*> * n_ptrLeavesList, double n_theta, 
					double n_epsilon, size_t n_groupSize = 32, double n_dt = 1,
					double n_G = 6.67259e-11 );
				void SetLeaves( std::vector<Leaf*> * n_ptrLeavesList );
				void Step();
				void Write(DataWriter * dw);

			protected:

				std::vector<Leaf*> * m_ptrLeavesList;
				std::vector<Group*> m_groupsList;

				double m_theta;
				double m_epsilon;
				size_t m_groupSize;
				double m_dt;
				double m_G;

			private:

				void CreateGroups(Node * node);

			};	//  end class Tree

		}	//  end namespace BarnesHut

	}	//  end namespace Simulations

}  //  end namespace JakubSchwarzmeier

#endif //  _JSSIMULATION_BH_H
