#include "../../ics/ext/inc/schwarzmeier/xmllib.h"

#include <fstream>
#include <iostream>

#include "../inc/simulation/direct.h"

using namespace Simulations::BarnesHut;

static const std::string s_strBarnesHut( "barneshut" );
static const std::string s_strTheta( "theta" );
static const std::string s_strEpsilon( "epsilon" );
static const std::string s_strGroupsize( "groupsize" );
static const std::string s_strTimestep( "timestep" );
static const std::string s_strGconst( "gconst" );

int main(int argc, char *argv[])
{
	if (argc < 3)  return -1;

	//  read parameters
	//

	xmlThrDefIndentTreeOutput( 1 );
	xmlKeepBlanksDefault(0);

	//  Read parameters from XML file
	xmlDoc * pXmlDoc = xmlParseFile( (const char *)argv[1] );
	xmlNode * pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  read count
	std::string strCount = GetXPathResponse( pXmlDoc, pRootNode, "./snapshot/count/text()" );
	int count = atoi( strCount.c_str() );

	//  read filename
	std::string strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  try to find Barnes-Hut node
	xmlNode * pBarnesHutNode = SelectSingleNode( pXmlDoc, pRootNode, s_strBarnesHut );
	double theta, epsilon, timestep, gconst;
	int groupsize;
	if (pBarnesHutNode == NULL)
	{
		std::cout << "* using command-line parameters" << std::endl;

		//  add parameters
		pBarnesHutNode = xmlNewNode( NULL, (const xmlChar *)s_strBarnesHut.c_str() );
		xmlAddChild( pRootNode, pBarnesHutNode );

		theta = atof( argv[3] );
		epsilon = atof( argv[4] );
		groupsize = atoi( argv[5] );
		timestep = atof( argv[6] );
		gconst = atof( argv[7] );

		SetAttribute( pBarnesHutNode, "theta", theta );
		SetAttribute( pBarnesHutNode, "epsilon", epsilon );
		SetAttribute( pBarnesHutNode, "groupsize", groupsize );
		SetAttribute( pBarnesHutNode, "timestep", timestep );
		SetAttribute( pBarnesHutNode, "gconst", gconst );
	}
	else
	{
		std::cout << "* using parameters from XML" << std::endl;

		//  read parameters
		GetAttribute( pBarnesHutNode, s_strTheta, &theta );
		GetAttribute( pBarnesHutNode, s_strEpsilon, &epsilon );
		GetAttribute( pBarnesHutNode, s_strGroupsize, &groupsize );
		GetAttribute( pBarnesHutNode, s_strTimestep, &timestep );
		GetAttribute( pBarnesHutNode, s_strGconst, &gconst );
	}

	//  create empty tree
	Direct direct( gconst, timestep );

	//  read ICs from from binary file into tree
	std::vector<Leaf*> * leaves = new std::vector<Leaf*>;

	//  input/output binary stream
	std::fstream binInOut;
	binInOut.open(strFilename.c_str(), std::ios_base::in | std::ios_base::out | std::ios_base::binary );

	//  get last snapshot
	int startPos, stopPos, id;
	xmlNode * snapNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );
	GetAttribute( snapNode, "start", &startPos );
	GetAttribute( snapNode, "end", &stopPos );
	GetAttribute( snapNode, "id", &id );

	int seekPos = startPos;

	//  seek to last snapshot position
	binInOut.seekg( seekPos, std::ios_base::beg );

	DataReader dr;
	dr.OpenStream( &binInOut );

	while (seekPos < stopPos)
	{
		//  create leaf and inserts it between other leaves
		Leaf * leaf = new Leaf;
		leaf->Read(&dr);
		leaves->push_back( leaf );

		seekPos = binInOut.tellg();
	}

	direct.SetLeaves( leaves );
	
	//  perform computation
	DataWriter dw;
	dw.OpenStream( &binInOut );

	long nSteps = atol( argv[2] );
	for(long i = id; i < id+nSteps; i++)
	{
		direct.Step();
		direct.Write( &dw );

		std::cout << "step " << i << std::endl;

		//  create new node
		xmlNode * pSnapshotNode = xmlNewNode( NULL, (const xmlChar *)"snapshot" );
		SetAttribute( pSnapshotNode, "id", i + 1 );
		SetAttribute( pSnapshotNode, "start", static_cast<long long>(count*sizeof(double)*7*(i+1)) );
		SetAttribute( pSnapshotNode, "end", static_cast<long long>( (count*sizeof(double)*7*(i+2)) - 1) );

		xmlNode * pCountNode = xmlNewNode( NULL, (const xmlChar *)"count" );
		SetTextNode( pXmlDoc, pCountNode, count );
		xmlAddChild( pSnapshotNode, pCountNode );

		xmlAddChild( pRootNode, pSnapshotNode );

		xmlSaveFormatFile (argv[1], pXmlDoc, 1);
	}

	xmlFreeDoc(pXmlDoc);

	binInOut.close();

	std::cout << "AMON finished.\a" << std::endl;

	return 0;
}

