/************************************************************************
 *
 *  main.cpp:  main C++ program for N-body simulation
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../../ics/ext/inc/schwarzmeier/xmllib.h"

#include <fstream>
#include <iostream>
#include <time.h>

#include "../inc/simulation/bh.h"
#include "../inc/simulation/direct.h"

#ifdef _MSC_VER
	#include "windows.h"
#endif


using namespace JakubSchwarzmeier::Simulations::BarnesHut;

static const std::string s_strNBody( "nbody" );
static const std::string s_strCode( "code" );
static const std::string s_strDirect( "direct" );
static const std::string s_strBarnesHut( "barneshut" );
static const std::string s_strTheta( "theta" );
static const std::string s_strEpsilon( "epsilon" );
static const std::string s_strGroupsize( "groupsize" );
static const std::string s_strTimestep( "timestep" );
static const std::string s_strGconst( "gconst" );

int main(int argc, char *argv[])
{
	if (argc < 3)  return -1;

	//  read parameters
	//

	xmlThrDefIndentTreeOutput( 1 );
	xmlKeepBlanksDefault(0);

	//  Read parameters from XML file
	xmlDoc * pXmlDoc = xmlParseFile( (const char *)argv[2] );
	xmlNode * pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  read count
	std::string strCount = GetXPathResponse( pXmlDoc, pRootNode, "./snapshot/count/text()" );
	long long count = atof( strCount.c_str() );

	//  read filename
	std::string strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  try to find Barnes-Hut node
	xmlNode * pNbodyNode = SelectSingleNode( pXmlDoc, pRootNode, s_strNBody );
	double theta, epsilon, timestep, gconst;
	int groupsize;
	if (pNbodyNode == NULL)
	{
		std::cout << "* using command-line parameters" << std::endl;

		//  add parameters
		pNbodyNode = xmlNewNode( NULL, (const xmlChar *)s_strNBody.c_str() );
		xmlAddChild( pRootNode, pNbodyNode );

		theta = atof( argv[4] );
		epsilon = atof( argv[5] );
		groupsize = atoi( argv[6] );
		timestep = atof( argv[7] );
		gconst = atof( argv[8] );

		std::string strVarCode;
		if ( tolower(*argv[1]) == 'b' )	
		{
			strVarCode = s_strBarnesHut;

			SetAttribute( pNbodyNode, "theta", theta );
			SetAttribute( pNbodyNode, "groupsize", groupsize );
		}
		else
			strVarCode = s_strDirect;

		xmlNewProp( pNbodyNode, (const xmlChar *)s_strCode.c_str(), (const xmlChar *)strVarCode.c_str() );

		SetAttribute( pNbodyNode, "epsilon", epsilon );
		SetAttribute( pNbodyNode, "timestep", timestep );
		SetAttribute( pNbodyNode, "gconst", gconst );
	}
	else
	{
		std::cout << "* using parameters from XML" << std::endl;

		//  read parameters
		GetAttribute( pNbodyNode, s_strTheta, &theta );
		GetAttribute( pNbodyNode, s_strEpsilon, &epsilon );
		GetAttribute( pNbodyNode, s_strGroupsize, &groupsize );
		GetAttribute( pNbodyNode, s_strTimestep, &timestep );
		GetAttribute( pNbodyNode, s_strGconst, &gconst );
	}

	//  create empty tree
	Tree bh(NULL, theta, epsilon, groupsize, timestep, gconst );
	Direct direct( gconst, timestep );

	//  read ICs from from binary file into tree
	std::vector<Leaf*> * leaves = new std::vector<Leaf*>;

	//  input/output binary stream
	JakubSchwarzmeier::Simulations::BarnesHut::fstream binInOut;
	binInOut.open( strFilename.c_str(), std::ios_base::in | std::ios_base::out | std::ios_base::binary );

	//  get last snapshot
	long long startPos, stopPos, id;
	xmlNode * snapNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );
	GetAttribute( snapNode, "start", &startPos );
	GetAttribute( snapNode, "end", &stopPos );
	GetAttribute( snapNode, "id", &id );

	long long seekPos = startPos;

	//  seek to last snapshot position
	binInOut.seekg( seekPos, std::ios_base::beg );

	DataReader dr;
	dr.OpenStream( &binInOut );

	while (seekPos < stopPos)
	{
		//  create leaf and inserts it between other leaves
		Leaf * leaf = new Leaf;
		leaf->Read(&dr);
		leaves->push_back( leaf );

		seekPos = binInOut.tellg();
	}

	bh.SetLeaves( leaves );
	direct.SetLeaves( leaves );
	
	//  perform computation
	DataWriter dw;
	dw.OpenStream( &binInOut );

	#ifdef _MSC_VER
		SetPriorityClass(GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS);
	#endif

	#ifdef MOND
		std::cout << "Running in MONDified mode." << std::endl;
	#endif

	long long nSteps = atof( argv[3] );
	std::cout << "Required space: " << ( 7.0 * sizeof(double) * leaves->size() * nSteps ) / (1024.0 * 1024.0)  << " MB" << std::endl;
	for(long long i = id; i < id+nSteps; i++)
	{
		clock_t start = clock();
		if ( tolower(*argv[1]) == 'b' )	
			bh.Step();
		else
			direct.Step();

		clock_t finish = clock();
		if ( tolower(*argv[1]) == 'b' )	
			bh.Write( &dw );
		else
			direct.Write( &dw );

		double duration = (double)(finish - start) / CLOCKS_PER_SEC;
		std::cout << "step " << i << ", " << duration << " s" << std::endl;

		if (i == id+1)
		{
			std::cout << "Estimated time: " << ((((duration*nSteps)/(60.0))/60.0)/24.0) << " days" << std::endl;
		}

		//  create new node
		xmlNode * pSnapshotNode = xmlNewNode( NULL, (const xmlChar *)"snapshot" );
		SetAttribute( pSnapshotNode, "id", static_cast<long long>(i + 1) );
		SetAttribute( pSnapshotNode, "start", static_cast<long long>(count*sizeof(double)*7*(i+1)) );
		SetAttribute( pSnapshotNode, "end", static_cast<long long>( (count*sizeof(double)*7*(i+2)) - 1) );

		xmlNode * pCountNode = xmlNewNode( NULL, (const xmlChar *)"count" );
		SetTextNode( pXmlDoc, pCountNode, count );
		xmlAddChild( pSnapshotNode, pCountNode );

		xmlAddChild( pRootNode, pSnapshotNode );

		xmlSaveFormatFile (argv[2], pXmlDoc, 1);
	}

	xmlFreeDoc(pXmlDoc);

	binInOut.close();

	std::cout << "AMON finished.\a" << std::endl;
	getchar();

	return 0;
}

