/*************************************************************************************
 *
 *  velocity.cpp:  extract rotation curve from binary file to gnuplot
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../../inc/reduction/velocity.h"

#include "../../../ics/ext/inc/schwarzmeier/xmllib.h"

#include <direct.h>

#include <limits>
#include <algorithm>
#include <functional>		//  std::binary_function

template <class _T, class _T2>
void GetSnapshot( long long no, _T * binIn, _T2 * dr,
				 Reduction::vecLeaves * leaves, xmlDoc * pXmlDoc, 
				 xmlNode * pRootNode )
{
	char szXPathNode[128];
	sprintf(szXPathNode, "./snapshot[@id=%d]", no );
	xmlNode * pCurrentNode = SelectSingleNode( pXmlDoc, pRootNode, szXPathNode );

	long long startPos, stopPos;
	GetAttribute( pCurrentNode,  "start", &startPos );
	GetAttribute( pCurrentNode,  "end", &stopPos );

	long long seekPos = startPos;

	binIn->seekg( seekPos, std::ios_base::beg );

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf * leaf = 
			new JakubSchwarzmeier::Simulations::BarnesHut::Leaf;
		leaf->Read( dr );
		leaves->push_back( leaf );

		seekPos = binIn->tellg();
	}
}

class SmallerDistance: public std::binary_function< JakubSchwarzmeier::Simulations::BarnesHut::Leaf*,
	JakubSchwarzmeier::Simulations::BarnesHut::Leaf*, bool > 
{
public:
	bool operator()( JakubSchwarzmeier::Simulations::BarnesHut::Leaf * elem1, 
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf * elem2) const 
    {
		return elem1->m_position.GetLength() < elem2->m_position.GetLength();
    }
};

void Reduction::PlotVcurve( std::string inFileName, std::string outDirectory,
	unsigned int outFileTabs, bool invokePlot )
{
	//  Read parameters from XML file
	xmlDoc * pXmlDoc = xmlParseFile( (const char *)inFileName.c_str() );
	xmlNode * pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  get binary filename
	std::string strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	xmlNode * lastNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );

	long long count;
	GetAttribute( lastNode,  "id", &count );

	char szCount[64];

	#ifndef _MSC_VER
		sprintf(szCount, "%lld", count );
	#else
		sprintf(szCount, "%I64d", count );
	#endif
	size_t noLen = strlen( szCount );

	count++;

	_mkdir( outDirectory.c_str() );
	std::string animateFile = outDirectory + std::string("/animate.plt");
	std::ofstream animateStream;
	animateStream.open( animateFile.c_str() );
	animateStream << "set grid" << std::endl;
	animateStream << "unset key" << std::endl;
	animateStream << "set xlabel \"r\"" << std::endl;
	animateStream << "set ylabel \"v\"" << std::endl;
	animateStream << "load \"animate_range.plt\"" << std::endl << std::endl;

	JakubSchwarzmeier::Simulations::BarnesHut::fstream binIn;
	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	JakubSchwarzmeier::Simulations::BarnesHut::DataReader dr;
	dr.OpenStream( &binIn );

	//  read from binary file into tree
	vecLeaves leaves;

	double xmin = DBL_MAX;
	double xmax = 0.0;
	double ymin = DBL_MAX;
	double ymax = 0.0;

	std::cout << "Generating rotation curve: " << strFilename << std::endl;

	for(long long i = 0; i < count; i++)
	{
		if (count > 100)
			if ( i % (count / 100) == 0 )
				std::cout << i << " of " << count-1 << std::endl;

		GetSnapshot<JakubSchwarzmeier::Simulations::BarnesHut::fstream,
		JakubSchwarzmeier::Simulations::BarnesHut::DataReader>( i, 
			&binIn, &dr, &leaves, pXmlDoc, pRootNode );

		sort( leaves.begin( ), leaves.end( ), SmallerDistance() );

		//  write
		char szFormatString[256];
		sprintf(szFormatString, "%%0%dd", noLen );

		char szFileName[256];
		sprintf(szFileName, szFormatString, i );
		std::string outFile = outDirectory + "/plot_v(r)_" + szFileName + ".txt";

		std::ofstream plotStream;
		plotStream.open( outFile.c_str() );
		plotStream.setf(std::ios::scientific);

		size_t step = (leaves.size() / outFileTabs);
		if (leaves.size() < outFileTabs)	step = 1;
		for(size_t j = 0; j < leaves.size(); j += step )
		{
			plotStream << leaves.at(j)->m_position.GetLength() << " " << 
				leaves.at(j)->m_velocity.GetLength() << std::endl;

			//  find limits
			if ( leaves.at(j)->m_position.GetLength() > xmax )
				xmax = leaves.at(j)->m_position.GetLength();
			if ( leaves.at(j)->m_position.GetLength() < xmin )
				xmin = leaves.at(j)->m_position.GetLength();

			if ( leaves.at(j)->m_velocity.GetLength() > ymax )
				ymax = leaves.at(j)->m_velocity.GetLength();
			if ( leaves.at(j)->m_velocity.GetLength() < ymin )
				ymin = leaves.at(j)->m_velocity.GetLength();
		}

		vecLeaves::iterator it;
		for( it = leaves.begin(); it != leaves.end(); it++ )
		{
			delete (*it);
		}

		leaves.clear();

		std::string pngFile = std::string("animate_") + szFileName + std::string(".png");
		animateStream << "set terminal png" << std::endl;
		animateStream << "set output \"" << pngFile << "\"" << std::endl;
		animateStream << "set title \"" << i << "\"" << std::endl;
		animateStream << "plot \"plot_v(r)_" << szFileName << ".txt\" smooth bezier" << std::endl;
		animateStream << std::endl;
	}

	animateStream << "quit" << std::endl;

	std::ofstream rangeStream;
	std::string outRangeStr = outDirectory + std::string("/animate_range.plt");
	rangeStream.open( outRangeStr.c_str() );
	rangeStream << "set xrange [" << xmin << ":" << xmax << "]" << std::endl;
	rangeStream << "set yrange [" << ymin << ":" << ymax << "]" << std::endl;
	rangeStream << std::endl;
}

