/*************************************************************************************
 *
 *  bh.cpp:  implementation of Barnes-Hut algorithm
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

/*
*  Reference:
*
*	J. E. Barnes, P. Hut, Nature, 1986, vol. 324, p. 446.
*
*/

#include "../../inc/simulation/bh.h"

#include <stack>
#include <omp.h>

using namespace JakubSchwarzmeier::Simulations::BarnesHut;

Tree::Tree( std::vector<Leaf*> * n_ptrLeavesList, double n_theta, double n_epsilon, 
		   size_t n_groupSize/*= 32*/, double n_dt/*= 1*/, double n_G/*= 6.67259e-11*/ )
{
	SetLeaves(n_ptrLeavesList);
	m_theta = n_theta;
	m_epsilon = n_epsilon;
	m_groupSize = n_groupSize;
	m_dt = n_dt;
	m_G = n_G;
}

void Tree::SetLeaves( std::vector<Leaf*> * n_ptrLeavesList )
{
	m_ptrLeavesList = n_ptrLeavesList;
}

void Tree::CreateGroups(Node * node)
{
	if (node->m_leavesList.size() <= m_groupSize)
	{
		m_groupsList.push_back( new Group( node ) );
	}
	else
	{
		for(size_t pos = 0; pos < node->m_childList.size(); pos++)
		{
			if ( typeid( *(node->m_childList[pos]) ) == typeid( Leaf ) )
			{
           		m_groupsList.push_back( new Group( node->m_childList[pos] ) );
			}
			else
			{
				CreateGroups( dynamic_cast<Node*>(node->m_childList[pos]) );
			}
		}
	}
}

void Tree::Step()
{
	//  when no leaves, then nothing to do
	if (m_ptrLeavesList->empty())  return;

	//  Phase I
	Construct( m_ptrLeavesList );

	const std::vector<Leaf*> & n_leavesList = *m_ptrLeavesList;

	//  prepare acceleration for this step
	for(size_t i = 0; i < n_leavesList.size(); i++)
	{
		n_leavesList[i]->m_acceleration = 0;
	}
    
	//  Clear old costzones records
	for(size_t i = 0; i < n_leavesList.size(); i++)	n_leavesList[i]->m_load = 0;

	//  WD99 -- Create groups
	//
	Node * root = m_octTree[0]->at(0);	//  root
	CreateGroups(root);

	//
	//  end WD99 -- Create groups


	//  Phase II: computing vector "center of mass" and scalar "mass"
	//

	//  for all levels including root
	for(size_t level = m_octTree.size(); level > 0; level--)
	{
		for( size_t pos = 0; pos < m_octTree[level-1]->size(); pos++ )
		{
			Node * actual = m_octTree[level-1]->at(pos);
			Math::Vector<> numerator = 0;

			for(size_t i = 0; i < actual->m_childList.size(); i++)
			{
				actual->m_mass += actual->m_childList[i]->m_mass;
				//  multiplication of vector with number
				numerator += actual->m_childList[i]->m_mass * actual->m_childList[i]->m_position;
			}
			//  division of a vector by a number
			actual->m_position = numerator / actual->m_mass;	//  "center of mass" (geometric quantity)
		}

	}

	//
	//  end Phase II: computing vector "center of mass" and scalar "mass"

	//  Phase 3: force evaluation
	//

	#pragma omp parallel for
	for(int i = 0; i < m_groupsList.size(); i++)	
	{
		std::stack< size_t >::value_type pos = 0;
		Node * cell = root;

		std::stack<Node*> openedStack;
		std::stack<size_t> positionStack;

		Group * group = m_groupsList[i];
		
		bool done;
		
		do
		{
			while ( pos < cell->m_childList.size() )	//  walk through all children
			{
				Math::Vector<> r = cell->m_childList[pos]->m_position - group->m_element->m_position;
				double d = r.GetLength();

				if ( cell->m_childList[pos] == group->m_element ) { pos++; }	//  avoid group (leave/node) self-interactions (and 1/(r=0)=-1.#IND00000 )	A NEMEL BYCH KONTROLOVAT r? (nekdy se castice mohou dostat tak blizko a pak by to zkolabovalo!!!)
				else
				if ( typeid( *(cell->m_childList[pos]) ) == typeid( Leaf ) )
				{
					//  group is a leave so put it into nearList
					group->m_nearList.push_back( cell->m_childList[pos] );
					pos++;
				}
				else
				if (d * m_theta >= dynamic_cast<Node*>(cell->m_childList[pos])->m_bisect)
				{
					double sphere;
					if ( typeid(*(group->m_element)) == typeid( Leaf ) )  goto push;
					
					sphere = 3. * dynamic_cast<Node*>(group->m_element)->m_bisect * sqrt(3.) / 2.;
					if ( d > sphere )
					{
						#ifndef MOND
						group->m_acceleration += m_G * r * cell->m_childList[pos]->m_mass / pow( d+m_epsilon, 3 );  //  far list
						#else
						Math::Vector<> accNewtonian = m_G * r * cell->m_childList[pos]->m_mass / pow( d+m_epsilon, 3 );  //  far list
						group->m_acceleration += (accNewtonian / M_SQRT2) * sqrt( (1.0 + sqrt(1.0+pow(2*3.5e-19/accNewtonian.GetLength(), 2.0))) );
						#endif
						cell->m_childList[pos]->m_load++;	//  costzones
					}
					else
					{	//  group is a node that is too close to group
						push: group->m_nearList.push_back( cell->m_childList[pos] );
					}
					pos++;
				}
				else
				{
					openedStack.push( cell );
					positionStack.push( pos );

					cell = dynamic_cast<Node*>( cell->m_childList[pos] );
					pos = 0;
				}
			}
			
			if ( !openedStack.empty() )
			{
				cell = openedStack.top(); openedStack.pop();
				pos = positionStack.top(); positionStack.pop();
				pos++;
				done = false;
			}
			else
			{
				done = true;
			}

		} while (!done);

		//  finish acceleration computing, because of WD99

		if ( typeid( *group->m_element ) == typeid(Leaf) )	//  only one particle in cell
		{
			//  add far forces acceleration
			//
			dynamic_cast<Leaf*>(group->m_element)->m_acceleration += group->m_acceleration;

			//  compute near forces
			//

			for(size_t j = 0; j < group->m_nearList.size(); j++)
			{
				Math::Vector<> r = group->m_nearList[j]->m_position - group->m_element->m_position;

				#ifndef MOND
				dynamic_cast<Leaf*>(group->m_element)->m_acceleration += m_G * r * group->m_nearList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
				#else
				Math::Vector<> accNewtonian = m_G * r * group->m_nearList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
				dynamic_cast<Leaf*>(group->m_element)->m_acceleration += (accNewtonian / M_SQRT2) * sqrt( (1.0 + sqrt(1.0+pow(2*3.5e-19/accNewtonian.GetLength(), 2.0))) );
				#endif

				group->m_nearList[j]->m_load++;	//  costzones
			}
		}
		else	//  more particles in cell
		{
			for(size_t k = 0; k < dynamic_cast<Node*>(group->m_element)->m_leavesList.size(); k++)
			{
				//  for leave "k"			
				
				//  add far forces
				#ifndef MOND
				dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += group->m_acceleration;
				#else
				Math::Vector<> accNewtonian = group->m_acceleration;
				dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += (accNewtonian / M_SQRT2) * sqrt( (1.0 + sqrt(1.0+pow(2*3.5e-19/accNewtonian.GetLength(), 2.0))) );
				#endif

				size_t j;

				//  add near forces
				for(j = 0; j < group->m_nearList.size(); j++)
				{
					Math::Vector<> r = group->m_nearList[j]->m_position - dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_position;

					#ifndef MOND
					dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += m_G * r * group->m_nearList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
					#else
					Math::Vector<> accNewtonian = m_G * r * group->m_nearList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
					dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += (accNewtonian / M_SQRT2) * sqrt( (1.0 + sqrt(1.0+pow(2*3.5e-19/accNewtonian.GetLength(), 2.0))) );
					#endif

					group->m_nearList[j]->m_load++;	//  costzones
				}
				//  add acceleration from leaves in grouping cell
				for(j = 0; j < dynamic_cast<Node*>(group->m_element)->m_leavesList.size(); j++)
				{
					if (k != j)
					{
						Math::Vector<> r = dynamic_cast<Node*>(group->m_element)->m_leavesList[j]->m_position - dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_position;

						#ifndef MOND
						dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += m_G * r * dynamic_cast<Node*>(group->m_element)->m_leavesList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
						#else
						Math::Vector<> accNewtonian = m_G * r * dynamic_cast<Node*>(group->m_element)->m_leavesList[j]->m_mass / pow( r.GetLength() + m_epsilon, 3 );
						dynamic_cast<Node*>(group->m_element)->m_leavesList[k]->m_acceleration += (accNewtonian / M_SQRT2) * sqrt( (1.0 + sqrt(1.0+pow(2*3.5e-19/accNewtonian.GetLength(), 2.0))) );
						#endif

						dynamic_cast<Node*>(group->m_element)->m_leavesList[j]->m_load++;
					}
				}
			}
			
		}
		
	} //  evaluate all leaves

	//
	//  end phase 3: force evaluation

	//  integrate, leap-frog

	for(size_t i = 0; i < n_leavesList.size(); i++ )
	{
		n_leavesList[i]->m_velocity += n_leavesList[i]->m_acceleration * m_dt;
		n_leavesList[i]->m_position += n_leavesList[i]->m_velocity * m_dt;
	}
	
	//  propagate costzones load info to leaves
	for(size_t level = 0; level < m_octTree.size(); level++)
	{
		
		for(size_t pos = 0; pos < m_octTree[level]->size(); pos++)
		{
			for(size_t i = 0; i <  m_octTree[level]->at(pos)->m_childList.size(); i++)
			{
				m_octTree[level]->at(pos)->m_childList[i]->m_load += m_octTree[level]->at(pos)->m_load / m_octTree[level]->at(pos)->m_childList.size();
			}
		}

	}

	//  free allocated groups
	for(size_t i = 0; i < m_groupsList.size(); i++)
	{
		delete m_groupsList[i];
	}

	m_groupsList.clear();

	FreeTree();
}

void Tree::Write(DataWriter * dw)
{
	for(size_t i = 0; i < m_ptrLeavesList->size(); i++)
	{
		(*m_ptrLeavesList)[i]->Write(dw);
	}
}
