#include "StdAfx.h"
#include "Animation.h"
#include "visualization.h"
#include "digalexapp.h"

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Std::Graphics;

CAnimation::CAnimation(void)
{
	m_startSnapshot = 0;
	m_startCamera = new CCamera;

	m_internalCamera = NULL;
	m_pVideoWriter = NULL;
	m_fps = 30.0;
}

CAnimation::~CAnimation(void)
{
	delete m_startCamera;
}

std::vector<CChunkDesc*> * CAnimation::GetChunks()
{
	return &m_chunks;
}

double CAnimation::GetStartSnapshot()
{
	return m_startSnapshot;
}

void CAnimation::SetStartSnapshot(double startTime)
{
	m_startSnapshot = startTime;
}

CCamera * CAnimation::GetStartCamera()
{
	return m_startCamera;
}

void CAnimation::SetStartCamera( CCamera * startMatrix )
{
	delete m_startCamera;
	m_startCamera 
		= dynamic_cast<CCamera*>( startMatrix->GetClassInfo()->CreateObject() );
	m_startCamera->Assign( startMatrix );

}

void CAnimation::Serialize(xmlDoc * pXmlDoc, xmlNode * branch)
{
	//  animation
	xmlNode * pAnimationNode =  xmlNewNode( NULL, (const xmlChar *)s_strAnimation.c_str() );
	xmlAddChild( branch, pAnimationNode );

	//  fps
	xmlNode * fpsNode = xmlNewNode( NULL, (const xmlChar *)s_strFps.c_str() );
	xmlAddChild( pAnimationNode, fpsNode );
	SetTextNode( pXmlDoc, fpsNode, m_fps );

	//  video file
	xmlNode * outputVideoNode = xmlNewNode( NULL, (const xmlChar *)s_strOutputVideo.c_str() );
	xmlAddChild( pAnimationNode, outputVideoNode );
	SetTextNode( pXmlDoc, outputVideoNode, m_outputVideo.c_str() );

	//  start snapshot
	xmlNode * snapshotNode = xmlNewNode( NULL, (const xmlChar *)s_strStartSnapshot.c_str() );
	xmlAddChild( pAnimationNode, snapshotNode );
	SetTextNode( pXmlDoc, snapshotNode, m_startSnapshot );

	//  start matrix
	xmlNode * startMatrixNode = xmlNewNode( NULL, (const xmlChar *)s_strStartCamera.c_str() );
	xmlAddChild( pAnimationNode, startMatrixNode );
	m_startCamera->Serialize( pXmlDoc, startMatrixNode );
	
	//  chunks
	xmlNode * cameraChunksNode = xmlNewNode( NULL, (const xmlChar *)s_strCameraChunks.c_str() );
	xmlAddChild( pAnimationNode, cameraChunksNode );

	std::vector<CChunkDesc*>::const_iterator it;
	for ( it = m_chunks.begin( ) ; it != m_chunks.end( ) ; it++ )
	{
		(*it)->Serialize( pXmlDoc, cameraChunksNode );
	}	
}

void CAnimation::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	//  get animation node
	xmlNode * animationNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strAnimation );

	//  fps
	std::string strFPS = GetXPathResponse( pXmlDoc, animationNode, s_strFps + "/text()" );
	m_fps = _tstof( strFPS.c_str() );

	//  output file
	std::string strOutputVideo = GetXPathResponse( pXmlDoc, animationNode, s_strOutputVideo + "/text()" );
	SetOutputVideo( strOutputVideo );

	//  get start snapshot
	std::string xpathSnapshotStr(s_strStartSnapshot);
	xpathSnapshotStr += "/text()";
	std::string start = GetXPathResponse( pXmlDoc, animationNode, xpathSnapshotStr );
	m_startSnapshot = atoi( start.c_str() );

	//  get start matrix
	xmlNode * startmatNode = SelectSingleNode( pXmlDoc, animationNode, s_strStartCamera );
	delete m_startCamera;
	m_startCamera = CCamera::Deserialize( startmatNode ) ;
	m_startCamera->Deserialize( pXmlDoc, startmatNode );

	//  deserialize cameras chunks
	xmlNode * camerasNode = SelectSingleNode( pXmlDoc, animationNode, s_strCameraChunks );
	if (camerasNode != NULL)
	{
		xmlNodeSet * cameraNode = SelectNodes( pXmlDoc, camerasNode, s_strCameraChunk );
		if ( cameraNode != NULL)
		{
			for(int i = 0; i < cameraNode->nodeNr; i++)
			{
				//  create object of the right class according to XML record of class name
				CChunkDesc * chunk = new CChunkDesc;
				chunk->Deserialize( pXmlDoc, cameraNode->nodeTab[i] );
				m_chunks.push_back( chunk );
			}
		}
	}
}

void CAnimation::SetOutputVideo(std::string outputVideo)
{
	m_outputVideo = outputVideo;
}

std::string CAnimation::GetOutputVideo()
{
	return m_outputVideo;
}

void CAnimation::SetFps(double fps)
{
	m_fps = fps;
}

double CAnimation::GetFps()
{
	return m_fps;
}

size_t CAnimation::Create (CVisualization * vis, double fps, int w, int h)
{
	m_pVisualization = vis;

	m_internalCamera = new CCamera;

	m_fps = fps;
	m_width = w;
	m_height = h;

	wxDocChildFrame* chFrm = reinterpret_cast<wxDocChildFrame*>(vis->GetThreedCanvas()->GetParent());

	vis->GetThreedCanvas()->GetSize( &m_initWidth, &m_initHeight );

	wxSize glWinSize( m_width, m_height );
	wxSize chldFrameSize = chFrm->GetSize();
	wxSize viewSize = vis->GetThreedCanvas()->GetSize();
	wxSize diffSize = chldFrameSize - viewSize;

	chFrm->SetSize( glWinSize + diffSize );
	vis->GetThreedCanvas()->SetSize( glWinSize );

	//  Create video writer
	m_pVideoWriter = cvCreateVideoWriter( GetOutputVideo().c_str(), 
	#if defined(__WXMSW__)
		-1,	/* let the user choose video codec */
	#else
        CV_FOURCC('P','I','M','1'), /* MPEG-1 codec */
	#endif
		m_fps, cvSize(m_width, m_height) );

	//  compute overall number of frames needed for production of whole animation
	size_t nTotalFrames = 0;
	std::vector<CChunkDesc*>::const_iterator it;
	for(it = m_chunks.begin(); it != m_chunks.end(); it++)
	{
		nTotalFrames += (*it)->GetDuration();
	}

	nTotalFrames *= m_fps;

	//  set initial values for animation generation
	if ( m_chunks.empty() )
		return -1;
	m_iChunk = 0;
	m_iChunkPos = -1;

    return nTotalFrames;
}

void CAnimation::Finish()
{
	//  release video writer
	cvReleaseVideoWriter( &m_pVideoWriter );

	delete m_internalCamera;

	wxDocChildFrame* chFrm = reinterpret_cast<wxDocChildFrame*>(m_pVisualization->GetThreedCanvas()->GetParent());
	m_pVisualization->GetThreedCanvas()->SetSize( m_initWidth, m_initHeight );
	chFrm->Fit();
}

void CAnimation::Update()
{
	//  render frame
	//

	//  get current chunk
	CChunkDesc * chunk = m_chunks.at( m_iChunk );
	if ( m_iChunkPos > ( chunk->GetDuration() * m_fps ) )
	{
		m_iChunkPos = 0;
		m_iChunk++;

		if (m_iChunk >= m_chunks.size())
			return;		//  we should already finish
		else
		{
			chunk = m_chunks.at( m_iChunk );
		}
	}
	else
	{
		m_iChunkPos++;
	}

	if (chunk->GetDuration() == 0)	throw;
	//  total number of steps (frames) in this chunk
	double steps = chunk->GetDuration() * m_fps;

	//  set time
	int simulationTimeSpan;
	if ( m_iChunk == 0) 
		simulationTimeSpan = chunk->GetEndSnapshot() - m_startSnapshot;
	else
		simulationTimeSpan = chunk->GetEndSnapshot() - m_chunks.at( m_iChunk-1 )->GetEndSnapshot();
	double simulationTimeStep = simulationTimeSpan / (steps);
	
	int currentSimulationTime;
	if ( m_iChunk == 0) 
		currentSimulationTime = simulationTimeStep * (double)m_iChunkPos + m_startSnapshot;
	else
		currentSimulationTime = simulationTimeStep * (double)m_iChunkPos + chunk->GetEndSnapshot();

	::wxGetApp().GetMainFrame()->SetSnapshot( currentSimulationTime );

	//  set camera
	CCamera * startCamera;
	if ( m_iChunk == 0) 
		startCamera = m_startCamera;
	else
		startCamera = m_chunks.at( m_iChunk-1 )->GetEndCamera();

	CCamera * endCamera = chunk->GetEndCamera();

	double t = (double)m_iChunkPos / steps;

	RotationMatrix startOrbiter = *startCamera;
	RotationMatrix endOrbiter = *endCamera;

	Vector<> direction;
	direction[0] = endOrbiter[12] - startOrbiter[12];
	direction[1] = endOrbiter[13] - startOrbiter[13];
	direction[2] = endOrbiter[14] - startOrbiter[14];

	Vector<> start;
	start[0] = startOrbiter[12];
	start[1] = startOrbiter[13];
	start[2] = startOrbiter[14];

	//  linear interpolation
	Vector<> currentPosition = start + t * direction;
	Quaternion<> currentOrientation = Slerp( startOrbiter.GetQuaternion(),
		endOrbiter.GetQuaternion(), t );

	RotationMatrix currentOrbiter(currentPosition, currentOrientation);

	JakubSchwarzmeier::wx::CObserver * obs 
		= m_pVisualization->GetThreedCanvas()->GetObserver();
	obs->SetCamera( currentOrbiter );


	m_pVisualization->GetThreedCanvas()->ForceReRenderWait();

	//  get the image
	m_pVisualization->GetThreedCanvas()->GetCopyImage( m_image );

	//  generate corresponding OpenCV image
	CvSize size;
    size.width = m_image.GetWidth();
    size.height = m_image.GetHeight();

	if ( (m_width != size.width) || (m_height != size.height) )
	{
		throw;
	}

	int depth = IPL_DEPTH_8U;

    IplImage * i_image = cvCreateImage( size, depth, 3 );
 
	//  Create continuous block of memory
	char * continuous = i_image->imageData; //  Intel allocated raw bitmap storage for us
	unsigned int vsize = 3;		//  3 bytes for one pixel

	for(int i = 0; i < size.height; i++)
	{
		for(int j = 0; j < size.width; j++)
		{
			for(int k = 0, l = 2; k < vsize; k++, l--)
			{
				char * src = reinterpret_cast<char *>(m_image.GetData() + i * vsize * size.width + j * vsize + l );

				//  there is padding at the end of each row (widthStep - width)
				memcpy( /*dest*/continuous + i * i_image->width * vsize + i * (i_image->widthStep/vsize - i_image->width) + j * vsize + k, 
					/*src*/src, vsize);
			}
		}
	}

	//  write animation frame
	int e = cvWriteFrame( m_pVideoWriter, i_image );
	if (e != 1) throw;

	//  release OpenCV image
	cvReleaseImage(&i_image);
}

std::string CAnimation::s_strAnimation( _T("animation") );
std::string CAnimation::s_strCameraChunks( _T("camerachunks") );
std::string CAnimation::s_strCameraChunk( _T("camerachunk") );

std::string CAnimation::s_strStartSnapshot( _T("startsnapshot") );
std::string CAnimation::s_strStartCamera( _T("startmatrix") );

std::string CAnimation::s_strFps( _T("fps") );
std::string CAnimation::s_strOutputVideo( _T("outputVideo") );
