/*************************************************************************************
 *
 *  kepler.cpp:  keplerian parameters --> cartesian position, velocity
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "kepler.h"

#define _USE_MATH_DEFINES
#include "math.h"

CKepler::CKepler()
{
	m_a = 0.0;
	m_e = 0.0;

	m_G = 1.0;
	m_ms = 0.0;
	m_mp = 0.0;

	m_ta = 0.0;

	m_px = m_py = 0.0;
	m_vx = m_vy = 0.0;
}

void CKepler::Set_a(double a)
{
	m_a = a;
}

void CKepler::Set_e(double e)
{
	m_e = e;
}

double CKepler::Get_a()
{
	return m_a;
}

double CKepler::Get_e()
{
	return  m_e;
}

void CKepler::SetG(double G)
{
	m_G = G;
}

void CKepler::Set_ms(double ms)
{
	m_ms = ms;
}

void CKepler::Set_mp(double mp)
{
	m_mp = mp;
}

void CKepler::Set_ta(double ta)
{
	m_ta = ta;
}

double CKepler::Get_ta()
{
	return m_ta;
}

void CKepler::Generate(double a, double e, 
	double G, double ms, double mp, 
	double ta )
{
	Set_a( a );
	Set_e( e );

	SetG( G );
	Set_ms( ms );
	Set_mp( mp );

	Set_ta( ta );

	Generate();
}

void CKepler::Generate()
{
	//  position
	m_px = m_a * ( 1 - m_e*m_e ) / ( 1 + m_e );
	m_py = 0.0;

	//  auxiliary for velocity
	double T = sqrt( (4.0 * M_PI * M_PI * pow( m_a, 3.0 )) / (m_G * (m_ms + m_mp)) );
	double n = 2.0 * M_PI / T;
	
	// velocity
	m_vx = 0.0;
	m_vy = n * m_a * sqrt( ( 1 + m_e ) / ( 1 - m_e ) );

	if (m_ta == 0)	return;	//  no need for 1-body simulation

	//  1-body
	double t = 0;
	double dt = T/1000.0;	//  timestep

	while (t < T)
	{
		//  distance
		double d = sqrt(m_px*m_px + m_py*m_py);
		//  acceleration
		double acc = -m_G * m_ms/ pow(d, 2.0);

		//  leap-frog integration
		m_vx += acc * dt;
		m_vy += acc * dt;
		m_px += m_vx * dt;
		m_py += m_vy * dt;

		double ca = atan2(m_py, m_px) * 180.0 / M_PI;
		if (ca >= m_ta)	break;
		
		//  update current time
		t += dt;
	}

	return;
}

double CKepler::GetPx()
{
	return m_px;
}

double CKepler::GetPy()
{
	return m_py;
}

double CKepler::GetVx()
{
	return m_vx;
}
	
double CKepler::GetVy()
{
	return m_vy;
}