/*************************************************************************************
 *
 *  compositemodel.cpp:  create galaxy from various components
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../inc/ics/compositemodel.h"

#include "../../inc/ics/plummermodel.h"

#include "../../inc/ics/flattenedmodel.h"

#include <algorithm>

CCompositeModel::CCompositeModel()
{
	m_name = "CompositeModel";

	m_rhor = NULL;
	m_accRhor = gsl_interp_accel_alloc();
}

CCompositeModel::~CCompositeModel()
{
	gsl_interp_accel_free (m_accRhor);
	if (m_rhor!=NULL)	gsl_spline_free( m_rhor );
}

void CCompositeModel::AddModel( CSphericalModel * model )
{
	m_models.push_back( model );
}

double CCompositeModel::rho(double r)
{
	return gsl_spline_eval (m_rhor, r, m_accRhor);
}

void CCompositeModel::Generate()
{
	std::vector<CSphericalModel*>::iterator iter;

	CInitialConditions::Generate();

	//--  Sum mass densities of all models

	//  allocate memory for grid of values
	double * rho = new double[m_funcTabs];
	double * radii = new double[m_funcTabs];

	//  create global mass density rho_total = rho_disk + ... + rho_halo
	const double max = TIMES_RADII * m_rMax;	//  out to 10x radii
	double d = 1.0;
	const double dStep = 1.0 / m_funcTabs;
	for(unsigned int i = 0; i < m_funcTabs; i++)
	{ 
		double r = CSphericalModel::log(d) * max;

		rho[i] = 0;
		//  loop through all bodies and recover their rho at position r
		for ( iter = m_models.begin() ; iter != m_models.end() ; iter++ )
		{
			CFlattenedModel * fm = dynamic_cast<CFlattenedModel*>((*iter));
			if (fm != NULL)
			{
                continue;                
			}

			rho[i] += (*iter)->rho( r );
		}

		radii[i] = r;

		d -= dStep;
	}

	//  Construct spline rho(r).
	m_rhor = gsl_spline_alloc(gsl_interp_cspline, m_funcTabs); 
	gsl_spline_init( m_rhor, radii, rho, m_funcTabs );

	delete [] rho;
	delete [] radii;

	//  Plot rho(r) if requested.
	if (m_bUsePlot)
	{
		m_plotOut.StartPlot( "r", "rho" );

		double plotStep = TIMES_RADII * m_rMax / m_plotTabs;
		for (unsigned int i = 0; i < m_plotTabs; i++ )
		{
			double r = i * plotStep;

			double rhox = CCompositeModel::rho( r );
			m_plotOut.Plot( r, rhox );
		}

		m_plotOut.StopPlot();
	}

	//  Compute global potential
	TabulateCumulativeMass();
	TabulatePotential();

	//  create arrays with global potential
	radii = new double[m_funcTabs];
	double * potential = new double[m_funcTabs];

	d = 1.0;
	for (unsigned int i = 0; i < m_funcTabs; i++ )
	{
		double r = log(d) * max;

		radii[i] = r;
		potential[i] = Potential(r);

		d -= dStep;
	}

    //  Insert global potential into all external models
	for ( iter = m_models.begin() ; iter != m_models.end() ; iter++ )
	{
		CFlattenedModel * fm = dynamic_cast<CFlattenedModel*>((*iter));
		if (fm != NULL)
		{
            continue;                
		}

		//  first create a copy of global potential
		gsl_spline * phir  = gsl_spline_alloc(gsl_interp_cspline, m_funcTabs); 
		gsl_spline_init( phir, radii, potential, m_funcTabs );

		//  insert the potential into model
		(*iter)->SetPotential( phir );
	}

	delete [] radii;
	delete [] potential;

	//  call rest of functions to finish individual model creation
	for ( iter = m_models.begin() ; iter != m_models.end() ; iter++ )
	{
		(*iter)->TabulateCumulativeMass();

		CFlattenedModel * fm = dynamic_cast<CFlattenedModel*>((*iter));
		if (fm != NULL)
		{
            (*iter)->TabulatePotential();
		}			

		(*iter)->Tabulate_rho_phi();
		(*iter)->TabulateDistributionFunction();
		(*iter)->Tabulate_rm();

		(*iter)->RealizeModel();
	}

	//  merge files
	std::string file = StartBinary();

	for ( iter = m_models.begin() ; iter != m_models.end() ; iter++ )
	{
		std::string inFile = (*iter)->GetOutputName();
		inFile += CInitialConditions::s_simNameExt;

		std::fstream binIn;
		binIn.open( inFile.c_str(), std::ios_base::in | std::ios_base::binary );
		m_fOut << binIn.rdbuf();
	}

	StopBinary();


	//  create XML
	m_count = 0;
	m_totalMass = 0;
	for ( iter = m_models.begin() ; iter != m_models.end() ; iter++ )
	{
		m_count += (*iter)->GetCount();
		m_totalMass += (*iter)->GetTotalMass();
	}

	xmlNode * pRootNode = StartXml();
	//  insert something into XML tree if needed
	StopXml(pRootNode);
}
