/*************************************************************************************
 *
 *  directcollision.cpp:  set system for direct collision on line
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../../inc/ics/directcollision.h"
#include "../../../../amon/inc/simulation/bh.h"

CDirectCollision::CDirectCollision(void)
{
	m_name = "DirectCollision";

	m_zDistance = 0.0;
	m_zVelocity = 0.0;
	m_xOffCenter = 0.0;

	m_px = m_py = m_pz = m_vx = m_vy = m_vz = 0.0;
	m_bColl2 = false;
}

CDirectCollision::~CDirectCollision(void)
{
}

void CDirectCollision::SetCentral( std::string fileName )
{
	m_centralXml = fileName;
}

void CDirectCollision::SetCompanion( std::string fileName )
{
	m_companionXml = fileName;
}

void CDirectCollision::SetCollision( double zDistance, double zVelocity, double xOffCenter )
{
	m_zDistance = zDistance;
	m_zVelocity = zVelocity;
	m_xOffCenter = xOffCenter;
}

void CDirectCollision::SetCollision2( double px, double py, double pz, 
	double vx, double vy, double vz )
{
	m_px = px;
	m_py = py;
	m_pz = pz;
	m_vx = vx;
	m_vy = vy;
	m_vz = vz;

	m_bColl2 = true;
}

void CDirectCollision::Generate()
{
	//  globals
	std::string strFilename;
	int startPos, stopPos;
	Math::Vector<> cm;

	xmlDoc * pXmlDoc;
	xmlNode * pRootNode;
	int seekPos;
	std::fstream binIn;
	JakubSchwarzmeier::Simulations::BarnesHut::DataReader dr;

	//  output binary stream
	std::string binICFile( m_outputName );
	binICFile += s_simNameExt;

	std::fstream binOut;
	binOut.open(binICFile.c_str(), std::ios_base::out | std::ios_base::binary );
	JakubSchwarzmeier::Simulations::BarnesHut::DataWriter dw;
	dw.OpenStream( &binOut );

	//-- Central Galaxy
	cm = CEncounter::GetCenterOfMass(m_centralXml);

	//  Read parameters from XML file
	pXmlDoc = xmlParseFile( (const char *)m_centralXml.c_str() );
	pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  get binary filename
	strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	xmlNode * lastNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );

	GetAttribute( lastNode,  "start", &startPos );
	GetAttribute( lastNode,  "end", &stopPos );

	//  position central galaxy according to galaxy's center of mass
	seekPos = startPos;

	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	binIn.seekg( seekPos, std::ios_base::beg );

	dr.OpenStream( &binIn );

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

		leaf.m_position = leaf.m_position /* - cm */;

		leaf.Write(&dw);

		m_count++;
		seekPos = binIn.tellg();
	}

	binIn.close();
	dr.close();
	
	//-- Companion Galaxy
	cm = CEncounter::GetCenterOfMass(m_companionXml);

	//  Read parameters from XML file
	pXmlDoc = xmlParseFile( (const char *)m_companionXml.c_str() );
	pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  get binary filename
	strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	lastNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );

	GetAttribute( lastNode,  "start", &startPos );
	GetAttribute( lastNode,  "end", &stopPos );

	//  position central galaxy according to galaxy's center of mass
	seekPos = startPos;

	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	binIn.seekg( seekPos, std::ios_base::beg );

	dr.OpenStream( &binIn );

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

		leaf.m_position = leaf.m_position /* - cm */;

		if (!m_bColl2)
		{
			leaf.m_position[0] += m_xOffCenter;
			leaf.m_position[2] += m_zDistance;
			leaf.m_velocity[2] += -m_zVelocity;
		}
		else
		{
			leaf.m_position[0] += m_px;
			leaf.m_position[1] += m_py;
			leaf.m_position[2] += m_pz;
			leaf.m_velocity[0] += m_vx;
			leaf.m_velocity[1] += m_vy;
			leaf.m_velocity[2] += m_vz;
		}

		leaf.Write(&dw);

		m_count++;
		seekPos = binIn.tellg();
	}

	//-- Output file
	xmlNode * pRootXmlNode = StartXml();

	xmlNode * pParametersNode = SelectSingleNode( m_pXmlDoc, pRootXmlNode, "./parameters" );

	xmlUnlinkNode( pParametersNode );
	xmlFreeNode( pParametersNode );

	StopXml(pRootXmlNode);
}
