/*************************************************************************************
 *
 *  keplerencounter.cpp:  set system for keplerian encounter
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../../inc/ics/keplerencounter.h"
#include "../../../../amon/inc/simulation/bh.h"

CKeplerEncounter::CKeplerEncounter()
{
	m_name = "KeplerEncounter";
}

void CKeplerEncounter::SetCentral( std::string fileName )
{
	m_centralXml = fileName;
}

void CKeplerEncounter::SetCompanion( std::string fileName )
{
	m_companionXml = fileName;
}

void CKeplerEncounter::SetCollision( double a, double e, double trueAnomaly )
{
	m_kepler.Set_a( a );
	m_kepler.Set_e( e );
	m_kepler.Set_ta( trueAnomaly );
}

void CKeplerEncounter::Generate()
{
	//  globals
	std::string strFilename;
	int startPos, stopPos;
	Math::Vector<> cm;

	xmlDoc * pXmlDoc;
	xmlNode * pRootNode;
	int seekPos;
	std::fstream binIn;
	JakubSchwarzmeier::Simulations::BarnesHut::DataReader dr;

	//  output binary stream
	std::string binICFile( m_outputName );
	binICFile += ".bin";

	std::fstream binOut;
	binOut.open(binICFile.c_str(), std::ios_base::out | std::ios_base::binary );
	JakubSchwarzmeier::Simulations::BarnesHut::DataWriter dw;
	dw.OpenStream( &binOut );

	//-- Central Galaxy
	cm = CEncounter::GetCenterOfMass(m_centralXml);

	//  Read parameters from XML file
	pXmlDoc = xmlParseFile( (const char *)m_centralXml.c_str() );
	pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  get binary filename
	strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	xmlNode * lastNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );

	GetAttribute( lastNode,  "start", &startPos );
	GetAttribute( lastNode,  "end", &stopPos );

	//  position central galaxy according to galaxy's center of mass
	seekPos = startPos;

	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	binIn.seekg( seekPos, std::ios_base::beg );

	dr.OpenStream( &binIn );

	double centralMass = 0.0;

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

//		leaf.m_position = leaf.m_position - cm;
		
		leaf.m_position[0] += m_kepler.Get_e() * m_kepler.Get_a();	//  add length eccentricity

		leaf.Write(&dw);

		centralMass += leaf.m_mass;

		m_count++;
		seekPos = binIn.tellg();
	}

	binIn.close();
	dr.close();

	//-- Companion Galaxy
	cm = CEncounter::GetCenterOfMass(m_companionXml);

	//  Read parameters from XML file
	pXmlDoc = xmlParseFile( (const char *)m_companionXml.c_str() );
	pRootNode = xmlDocGetRootElement(pXmlDoc);

	//  get binary filename
	strFilename = GetXPathResponse( pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	lastNode = SelectSingleNode( pXmlDoc, pRootNode, "./snapshot[last()]" );

	GetAttribute( lastNode,  "start", &startPos );
	GetAttribute( lastNode,  "end", &stopPos );

	//  position central galaxy according to galaxy's center of mass
	seekPos = startPos;

	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	binIn.seekg( seekPos, std::ios_base::beg );

	dr.OpenStream( &binIn );

	double companionMass = 0.0;
	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

		companionMass += leaf.m_mass;

		seekPos = binIn.tellg();
	}

	m_kepler.Set_ms(centralMass);
	m_kepler.Set_mp(companionMass);
	m_kepler.Generate();

	seekPos = startPos;
	binIn.seekg( seekPos, std::ios_base::beg );

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

		leaf.m_position[0] = leaf.m_position[0] - cm[0] + m_kepler.GetPx();
		leaf.m_position[1] = leaf.m_position[1] - cm[1] - m_kepler.GetPy();

		leaf.m_velocity[0] += m_kepler.GetVx();
		leaf.m_velocity[1] += m_kepler.GetVy();

		leaf.Write(&dw);

		m_count++;
		seekPos = binIn.tellg();
	}

	//-- Output file
	xmlNode * pRootXmlNode = StartXml();

	xmlNode * pParametersNode = SelectSingleNode( m_pXmlDoc, pRootXmlNode, "./parameters" );

	xmlUnlinkNode( pParametersNode );
	xmlFreeNode( pParametersNode );

	StopXml(pRootXmlNode);
}
