/*************************************************************************************
 *
 *  keplermodel.cpp:  generate Kepler's model
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../inc/ics/keplermodel.h"

#include "../../ext/src/astronomy/kepler.h"
#include "../../inc/ics/sphericalmodel.h"

CKeplerModel::CKeplerModel()
{
	m_name = "KeplerModel";
	m_centralMass = m_totalMass - (m_totalMass/1e3);
}

void CKeplerModel::Generate()
{
	CInitialConditions::Generate();
	
	CKepler kepler;
	kepler.SetG( m_G );
	kepler.Set_ms( m_centralMass );
	
	if (m_totalMass <= m_centralMass) throw;
	double bodyMass = (m_totalMass - m_centralMass) / (m_count-1);
	kepler.Set_mp( bodyMass );

	//  test energies
	double Epot = 0;
    double Ekin = 0;

	//  prepare uniform (normal, Gaussian) random number generator
	gsl_rng * rnd = gsl_rng_alloc ( gsl_rng_mt19937 );

	std::string file = StartBinary();

	for( unsigned int i = 0; i < m_count; i++)
	{
		double radius, mass, px, py, pz, vx, vy, vz;

		if (i == 0)
		{
			mass = m_centralMass;
			px = py = pz = 0;
			vx = vy = vz = 0;
		}
		else
		{
			mass = bodyMass;

			//  closest particle will be at m_rMax/100.0 to center
			radius =  m_rMax/100.0 + gsl_rng_uniform_pos( rnd ) * (m_rMax - m_rMax/100.0);

			Epot += (-m_G * m_centralMass / radius);

			kepler.Set_a( radius );

			double phi = 0 + gsl_rng_uniform_pos( rnd ) * ( 2 * M_PI + 0.0 );

			px = radius * cos( phi );
			py = radius * sin( phi );
			pz = 0;	

			kepler.Generate();
			double velocity = sqrt( pow( kepler.GetVx(), 2 ) + pow(kepler.GetVy(), 2) );
			vx = -velocity * sin( phi );
			vy = velocity * cos( phi );
			vz = 0;
		}

		BinaryWrite( mass );

		BinaryWrite( px );
		BinaryWrite( py );
		BinaryWrite( pz );

		BinaryWrite( vx );
		BinaryWrite( vy );
		BinaryWrite( vz );

		Ekin += 0.5 * ( vx * vx + vy * vy + vz * vz );
	}

	StopBinary();

    Epot *= 0.5;

	std::cout << "Ekin: " << Ekin << ", Epot: " << Epot << std::endl;
	std::cout << "virial ratio = " << -2.0*Ekin/Epot << std::endl;

	xmlNode * pRootNode = StartXml();
	//  insert something into XML tree if needed
	StopXml(pRootNode);

	//  free random number generator object
	gsl_rng_free( rnd );
}

void CKeplerModel::SetCentralMass(double centralMass)
{
	m_centralMass = centralMass;
}
