/*************************************************************************************
 *
 *  initialconditions.h:  base class for initial conditions
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../inc/ics/initialconditions.h"

CInitialConditions::CInitialConditions()
{
	m_name = "InitialConditions";

	m_rMax = 1;
	m_rCut = TIMES_RADII * m_rMax;
	m_totalMass = 1;
	m_G = 1.0;
	m_a = 1.0;

	m_outputName = m_name;
	m_bUsePlot = false;
	m_plotTabs = 500;

	m_count = 0;
	m_funcTabs = 1000;
}

CInitialConditions::~CInitialConditions(void)
{
}

void CInitialConditions::Generate()
{
	std::cout << "Generating " << m_name << std::endl;
}

void CInitialConditions::SetCount( unsigned int count )
{
	m_count = count;
}

unsigned int CInitialConditions::GetCount( )
{
	return m_count;
}

void CInitialConditions::SetFuncTabs( unsigned int funcTabs )
{
	m_funcTabs = funcTabs;
}

unsigned int CInitialConditions::GetFuncTabs( )
{
	return m_funcTabs;
}

void CInitialConditions::Set_r(double r)
{
	m_rMax = r;
}

double CInitialConditions::Get_rMax()
{
	return m_rMax;
}

void CInitialConditions::SetTotalMass(double massTotal)
{
	m_totalMass = massTotal;
}

double CInitialConditions::GetTotalMass()
{
	return m_totalMass;
}

void CInitialConditions::Set_a(double a)
{
	m_a = a;
}

double CInitialConditions::Get_a()
{
	return m_a;
}

void CInitialConditions::SetOutputName( std::string outputName )
{
	m_outputName = outputName;
}

std::string CInitialConditions::GetOutputName( )
{
	return m_outputName;
}

void CInitialConditions::SetPlot( bool bUsePlot )
{
	m_bUsePlot = bUsePlot;

	if (m_bUsePlot)
		m_plotOut.Create( m_outputName );
}

void CInitialConditions::SetPlotTabs( unsigned int plotTabs )
{
	m_plotTabs = plotTabs;
}

std::string CInitialConditions::StartBinary()
{
	std::string file( m_outputName );
	file += s_simNameExt;
	m_fOut.open( file.c_str(), std::ios::binary );

	return file;
}

void CInitialConditions::BinaryWrite(double val)
{
	m_fOut.write( reinterpret_cast<char*>(&val), sizeof(val) );
}

void CInitialConditions::StopBinary()
{
	m_fOut.close();
}

xmlNode * CInitialConditions::StartXml()
{
	//  Write XML file
	m_pXmlDoc = xmlNewDoc( (const xmlChar *)"1.0" );
	
	//  create new node
	xmlNode * pRootNode = xmlNewNode( NULL, (const xmlChar *)s_simulation.c_str() );

	//  connect root node to document
	xmlDocSetRootElement( m_pXmlDoc, pRootNode );

	xmlNewProp( pRootNode, (const xmlChar *)"class", (const xmlChar *)m_name.c_str() );

	//  create parameters node
	xmlNode * pParametersNode = xmlNewNode( NULL, (const xmlChar *)"parameters" );
	SetAttribute( pParametersNode, "m", m_totalMass );
	SetAttribute( pParametersNode, "r", m_rMax );
	SetAttribute( pParametersNode, "rcut", m_rCut );
	SetAttribute( pParametersNode, "a", m_a );

	xmlAddChild( pRootNode, pParametersNode );

	//  create units node
	xmlNode * pUnitsNode = xmlNewNode( NULL, (const xmlChar *)"units" );
	SetAttribute( pUnitsNode, "G", 1.0 );
	SetAttribute( pUnitsNode, "M", 1.0 );
	SetAttribute( pUnitsNode, "a", 1.0 );
	
	xmlAddChild( pRootNode, pUnitsNode );

	return pRootNode;
}

void CInitialConditions::StopXml(xmlNode * pRootNode)
{
	std::string file( m_outputName );
	file += s_simNameExt;

	//  create snapshot node
	xmlNode * pSnapshotNode = xmlNewNode( NULL, (const xmlChar *)"snapshot" );
	xmlNewProp( pSnapshotNode, (const xmlChar *)"id", (const xmlChar *)"0" );
	xmlNewProp( pSnapshotNode, (const xmlChar *)"start", (const xmlChar *)"0" );
	SetAttribute( pSnapshotNode, "end", (int)(m_count*sizeof(double)*7) - 1 );
	xmlAddChild( pRootNode, pSnapshotNode );

	//  create count node
	xmlNode * pCountNode = xmlNewNode( NULL, (const xmlChar *)"count" );
	SetTextNode( m_pXmlDoc, pCountNode, (int)m_count );
	xmlAddChild( pSnapshotNode, pCountNode );

	//  create filename node
	xmlNode * pFilenameNode = xmlNewNode( NULL, (const xmlChar *)"filename" );
	SetTextNode( m_pXmlDoc, pFilenameNode, file );

	//  plug new node into tree
	if ( xmlAddChild( pRootNode, pFilenameNode ) == NULL )
	{
		//  failure
		return;
	}

	//  save
	xmlThrDefIndentTreeOutput( 1 );
	xmlKeepBlanksDefault(0);

	std::string xmlFile( m_outputName );
	xmlFile += ".xml";
	xmlSaveFormatFile( xmlFile.c_str(), m_pXmlDoc, 1 );
}

void CInitialConditions::PrintFinish( double & Ekin, double & Epot )
{
    Epot *= 0.5;

	std::cout << "Ekin: " << Ekin << ", Epot: " << Epot << std::endl;
	std::cout << "virial ratio = " << -2.0*Ekin/Epot << std::endl;
	double vchar = ::sqrt(2.0*Ekin/m_totalMass);
	std::cout << "characteristic velocity = " << vchar << std::endl;
	double Rchar = (-m_G*pow(m_totalMass,2))/(2*Epot);
	std::cout << "characteristic length = " << Rchar << std::endl;
	std::cout << "characteristic time = " << (2 * Rchar) / vchar << std::endl;
}

const std::string CInitialConditions::s_simulation = "simulation";
const std::string CInitialConditions::s_simNameExt = "_ics.bin";
