/*************************************************************************************
 *
 *  rotate.cpp:  rotate galaxy around x, y, z axes
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include "../../inc/ics/rotate.h"

#include "../../../../amon/inc/simulation/bh.h"
#include "../../../../amon/inc/math/modelmatrix.h"

using namespace Math;

CRotate::CRotate()
{
	m_name = "UnitsConvertor";

	m_lambda = 0.0;
	m_phi = 0.0;
	m_theta = 0.0;
}

CRotate::~CRotate()
{
}

void CRotate::SetInputName( std::string input )
{
	m_input = input;
}

void CRotate::SetRotation( double lambda, double phi, double theta )
{
	m_lambda = lambda;
	m_phi = phi;
	m_theta = theta;
}

void CRotate::Generate()
{
	ModelMatrix mat;
	mat.LoadIdentity();
	mat.ChangeHeading(m_lambda);
	mat.ChangePitch(m_phi);
	mat.ChangeRoll(m_theta);

	xmlNode * pRootNode;
	long long seekPos;
	std::fstream binIn;
	JakubSchwarzmeier::Simulations::BarnesHut::DataReader dr;

	//  output binary stream
	std::string binICFile( m_outputName );
	binICFile += s_simNameExt;

	std::fstream binOut;
	binOut.open(binICFile.c_str(), std::ios_base::out | std::ios_base::binary );
	JakubSchwarzmeier::Simulations::BarnesHut::DataWriter dw;
	dw.OpenStream( &binOut );

	//  Read parameters from XML file
	m_pXmlDoc = xmlParseFile( (const char *)m_input.c_str() );
	pRootNode = xmlDocGetRootElement(m_pXmlDoc);

	//  get binary filename
	std::string strFilename = GetXPathResponse( m_pXmlDoc, pRootNode, "./filename/text()" );

	//  get last snapshot
	xmlNode * lastNode = SelectSingleNode( m_pXmlDoc, pRootNode, "./snapshot[last()]" );

	long long startPos, stopPos;
	GetAttribute( lastNode,  "start", &startPos );
	GetAttribute( lastNode,  "end", &stopPos );

	//  modify out filename
	std::string outFileNameBin(m_outputName);
	outFileNameBin += ".bin";
	xmlNode * fn = SelectSingleNode( m_pXmlDoc, pRootNode, "./filename/text()" );

	xmlNodeSetContent(fn, (const xmlChar *) outFileNameBin.c_str() );

	seekPos = startPos;

	binIn.open(strFilename.c_str(), std::ios_base::in | std::ios_base::binary );
	binIn.seekg( seekPos, std::ios_base::beg );

	dr.OpenStream( &binIn );

	while (seekPos < stopPos)
	{
		JakubSchwarzmeier::Simulations::BarnesHut::Leaf leaf;
		leaf.Read(&dr);

		ModelMatrix position;
		position.SetPosition( leaf.m_position );
		ModelMatrix velocity;
		velocity.SetPosition( leaf.m_velocity );
		for(int kk = 12; kk < 15; kk++)
		{
			position[kk] = -position[kk];
			velocity[kk] = -velocity[kk];
		}

		ModelMatrix res;
		res.LoadIdentity();
		res = position * mat;
		leaf.m_position[0] = res[12];
		leaf.m_position[1] = res[13];
		leaf.m_position[2] = res[14];

		res = mat * velocity;
		leaf.m_velocity[0] = res[12];
		leaf.m_velocity[1] = res[13];
		leaf.m_velocity[2] = res[14];

		leaf.Write(&dw);

		m_count++;
		seekPos = binIn.tellg();
	}

	binIn.close();
	dr.close();

	xmlNodeSet * pNodes = SelectNodes( m_pXmlDoc, pRootNode, "./snapshot" );

	//  clearing itself
	for(int i = 0; i < pNodes->nodeNr; i++)
	{
		xmlUnlinkNode(pNodes->nodeTab[i]);
		xmlFreeNode(pNodes->nodeTab[i]);
	}

	StopXml(pRootNode);
}
