/*************************************************************************************
 *
 *  main-composite.cpp:  defines the entry point for the composite model application
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include <vector>

#include "../../inc/ics/plummermodel.h"
#include "../../inc/ics/hernquistmodel.h"
#include "../../inc/ics/jaffemodel.h"
#include "../../inc/ics/isothermalsphere.h"
#include "../../inc/ics/uniformsphere.h"

#include "../../inc/ics/keplermodel.h"
#include "../../inc/ics/exponentialmodel.h"
#include "../../inc/ics/kuzminmodel.h"

#include "../../inc/ics/compositemodel.h"

void CmdParSetCount( CInitialConditions * model, const char * count )
{
	model->SetCount( atoi( count ) );
}

void CmdParSetOutputName( CInitialConditions * model, const char * outputName )
{
	model->SetOutputName( outputName );
}

void CmdParSetPlot( CInitialConditions * model, const char * bPlot )
{
	if ( tolower( *bPlot ) == 'p' )	model->SetPlot( true );
}

void help()
{
	std::cout << "Usage: icsmain model_code [model_parameter_1] [model_parameter_2] ... [model_parameter_N] output_file [p]" << std::endl;
	std::cout << "e.g. icsmain p 10000 ./bulge_component p" << std::endl;

	return;
}

int main(int argc, char *argv[])
{
	std::vector<CSphericalModel*> models;
	CCompositeModel composite;

    if (argc < 3)
	{
		//--  Probably a new user or wrong number of parameters. Print help message.
		help();
		return -1;
	}
	else
	{	
		int actualParam = 2;

		do
		{
			CSphericalModel * pNewModel;

			//--  Spheric or flat model?
			switch ( tolower( *argv[actualParam++] ) )
			{
			case 's':
			{
				//--  Choose a correct model.
				switch ( tolower( *argv[actualParam++] ) )
				{
				case 'h':
				{
					//--  Hernquist's model
					CHernquistModel * pHernquist = new CHernquistModel;

					CmdParSetOutputName( pHernquist, argv[actualParam++] );
					CmdParSetCount( pHernquist, argv[actualParam++] );
					pNewModel = pHernquist;
				}

					break;

				case 'p':
				{
					//--  Plummer's model
					CPlummerModel * pPlummer = new CPlummerModel;

					CmdParSetOutputName( pPlummer, argv[actualParam++] );
					CmdParSetCount( pPlummer, argv[actualParam++] );
					pNewModel = pPlummer;
				}

					break;

				case 'j':
				{
					//--  Jaffe's model
					CJaffeModel * pJaffe = new CJaffeModel;

					CmdParSetOutputName( pJaffe, argv[actualParam++] );
					CmdParSetCount( pJaffe, argv[actualParam++] );
					pNewModel = pJaffe;
				}

					break;

				case 't':
				{
					CIsothermalSphere * pIsoThermalSphere = new CIsothermalSphere;

					CmdParSetOutputName( pIsoThermalSphere, argv[actualParam++] );
					CmdParSetCount( pIsoThermalSphere, argv[actualParam++] );
					pIsoThermalSphere->SetParameters( atof( argv[actualParam++] ) );	//  0.036
					pNewModel = pIsoThermalSphere;
				}

					break;

				case 'u':
				{
					//--  Uniform sphere model
					CUniformSphere * pUniformSphere = new CUniformSphere;

					CmdParSetOutputName( pUniformSphere, argv[actualParam++] );
					CmdParSetCount( pUniformSphere, argv[actualParam++] );
					pNewModel = pUniformSphere;
				}

					break;

				default:

					//--  User entered wrong name of model
					help();
					return -1;
				}
			}

				break;

			case 'f':
			{
				//--  Choose a correct model.
				switch ( tolower( *argv[actualParam++] ) )
				{
				case 'k':
				{
					//--  Kepler's model
					std::cout << "Kepler model cannot be used in composite models." << std::endl;
				}
					break;

				case 'u':
				{
					//--  Kuzmin's model
					CKuzminModel * pKuzmin = new CKuzminModel;

					CmdParSetOutputName( pKuzmin, argv[actualParam++] );
					CmdParSetCount( pKuzmin, argv[actualParam++] );
					pNewModel = pKuzmin;
				}

					break;

				case 'e':
				{
					//--  Exponential model
					CExponentialModel * pExp = new CExponentialModel;

					CmdParSetOutputName( pExp, argv[actualParam++] );
					CmdParSetCount( pExp, argv[actualParam++] );
					pNewModel = pExp;
				}

					break;

				default:

					//--  User entered wrong name of model
					help();
					return -1;
				}
			}

				break;

			default:

				help();
				return -1;
			}

			models.push_back( pNewModel );


		} while (actualParam+1 < argc );

		std::vector<CSphericalModel*>::iterator iter;

		//  add components to composite model
        for ( iter = models.begin(); iter != models.end() ; iter++ )
		{
			composite.AddModel( (*iter) );
		}

		//  Generate model
		CmdParSetOutputName( &composite, argv[1] );
		if (actualParam<argc)
		{
			if ( tolower( *argv[actualParam] ) == 'p' )
			{
				for ( iter = models.begin(); iter != models.end() ; iter++ )
				{
					(*iter)->SetPlot( true );
				}				
				
				composite.SetPlot(true);
			}
		}
		composite.Generate();

		//  release allocated memory for models
		for ( iter = models.begin(); iter != models.end() ; iter++ )
		{
			delete (*iter);
		}
	}

	return 0;

}
