/*************************************************************************************
 *
 *  main-spheric.cpp:  defines the entry point for the spheric models application
 *
 *  Copyright (c) 2007 Jakub Schwarzmeier
 *  University of West Bohemia, Department of General Physics
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ************************************************************************/

#include "../includes.h"

#include <ics/plummermodel.h>
#include <ics/hernquistmodel.h>
#include <ics/jaffemodel.h>
#include <ics/isothermalsphere.h>
#include <ics/uniformsphere.h>

void CmdParSet( CInitialConditions * model, const char * outputName,
			   const char * count
			   )
{
	model->SetOutputName( outputName );
	model->SetCount( atoi( count ) );
	model->SetTotalMass( 1.0 );
	model->Set_r( 1.0 );
	model->Set_a( 1.0 );
}


void CmdParSetCount( CInitialConditions * model, const char * count )
{
	model->SetCount( atoi( count ) );
}

void CmdParSetOutputName( CInitialConditions * model, const char * outputName )
{
	model->SetOutputName( outputName );
}

void CmdParSetPlot( CInitialConditions * model, const char * bPlot )
{
	if ( tolower( *bPlot ) == 'p' )	model->SetPlot( true );
}

void help()
{
	std::cout << "Usage: icsmain model_code [model_parameter_1] [model_parameter_2] ... [model_parameter_N] output_file [p]" << std::endl;
	std::cout << "e.g. icsmain p 10000 ./bulge_component p" << std::endl;

	return;
}

int main(int argc, char *argv[])
{
    if (argc < 2)
	{
		//--  Probably a new user or wrong number of parameters. Print help message.
		help();
		return -1;
	}
	else
	{	
		//--  Choose a correct model.
		switch ( tolower( *argv[1] ) )
		{
		case 'h':
			if (argc > 3)
			{
				//--  Hernquist's model
				CHernquistModel hernquist;

				CmdParSet( &hernquist, argv[2], argv[3] );
				if (argc == 5)  CmdParSetPlot( &hernquist, argv[4] );
				hernquist.Generate();
			}
			else
			{
				help();
				return -1;
			}

			break;

		case 'p':
			//--  Plummer's model
			if (argc > 3)
			{
				CPlummerModel plummer;

				CmdParSetOutputName( &plummer, argv[2] );
				CmdParSetCount( &plummer, argv[3] );
				if (argc == 5)  CmdParSetPlot( &plummer, argv[4] );
				plummer.Generate();
			}
			else
			{
				help();
				return -1;
			}

			break;

		case 'j':
			//--  Jaffe's model
			if (argc > 3)
			{
				CJaffeModel jaffe;

				CmdParSetOutputName( &jaffe, argv[2] );
				CmdParSetCount( &jaffe, argv[3] );
				if (argc == 5)  CmdParSetPlot( &jaffe, argv[4] );
				jaffe.Generate();
			}
			else
			{
				help();
				return -1;
			}

			break;

		case 't':
			//--  Isothermal sphere model
			if (argc > 4)
			{
				CIsothermalSphere isoThermalSphere;

				CmdParSetOutputName( &isoThermalSphere, argv[2] );
				CmdParSetCount( &isoThermalSphere, argv[3] );
				isoThermalSphere.SetParameters( atof( argv[4] ) );	//  0.036
				if (argc == 6)  CmdParSetPlot( &isoThermalSphere, argv[5] );
				isoThermalSphere.Generate();
			}
			else
			{
				help();
				return -1;
			}

			break;

		case 'u':
			//--  Uniform sphere model
			if (argc > 3)
			{
				CUniformSphere uniformSphere;

				CmdParSetOutputName( &uniformSphere, argv[2] );
				CmdParSetCount( &uniformSphere, argv[3] );
				if (argc == 5)  CmdParSetPlot( &uniformSphere, argv[4] );
				uniformSphere.Generate();
			}
			else
			{
				help();
				return -1;
			}

			break;

		default:

			//--  User entered wrong name of model
			help();
			return -1;
		}

		return 0;
	}
}
